#include <cassert>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iostream>
#include <mpi.h>

#include <flens/flens.h>

#define rdtsc(val) __asm__ __volatile__ ("rdtsc" : "=A" (val))
    //------------------------------------------------------------------------------

    using namespace flens;
    using namespace std;


    //------------------------------------------------------------------------------

    enum CartDir { North=0, East=1, South=2, West=3};

    struct MpiCart
    {
        MpiCart()
        {}
        
        MpiCart(int _numRows, int _numCols)
            : numRows(_numRows), numCols(_numCols)
        {
            int  dims[2];
            bool periods[2];

            // create mpi_cart:  2-dimensional, 2 x 2, non-periodic
            dims[0]    = numRows;
            dims[1]    = numCols;
            periods[0] = false;
            periods[1] = false;
            comm = MPI::COMM_WORLD.Create_cart(2, dims, periods, true);
            
            // find your position in the cart
            rank = MPI::COMM_WORLD.Get_rank();
            int coords[2];
            comm.Get_coords(rank, 2, coords);
            row = coords[0];
            col = coords[1];
        
            // find your neighbors
            comm.Shift(0, 1, neighbors[North], neighbors[South]);
            comm.Shift(1, 1, neighbors[West],  neighbors[East]);
        }
        
        MpiCart(const MpiCart &rhs)
            : comm(rhs.comm),
              numRows(rhs.numRows), numCols(rhs.numCols),
              row(rhs.row), col(rhs.col)
        {
            for (int k=0; k<8; ++k) {
                neighbors[k] = rhs.neighbors[k];
            }
        }

        int
        cartRank(int row, int col)
        {
            if ((row<0) || (col<0) || (row==numRows) || (col==numCols)) {
                return MPI_PROC_NULL;
            }
            int coords[2];
            coords[0] = row;
            coords[1] = col;
            return comm.Get_cart_rank(coords);
        }
        
        MPI::Cartcomm comm;
        int numRows, numCols;
        int row, col;
        int rank;
        int neighbors[4];
    };

    struct DistributedVector
    {
        typedef GeMatrix<FullStorage<double, RowMajor> > Grid;

        DistributedVector()
        {
        }

        DistributedVector(const MpiCart &_mpiCart, int rh)
            : mpiCart(_mpiCart), N(rh-1)
        {
            // global matrix has size (N+1) x (N+1)
            
            // compute m and n, such that local matrix has size (-1,m+1) x (-1,n+1)
            m = (N+1) / mpiCart.numRows - 1;
            n = (N+1) / mpiCart.numCols - 1;
            grid.resize(_(-1,m+1),_(-1,n+1));
            
            // m0, n0: row and col offset
            m0 = (m+1)*mpiCart.row;
            n0 = (n+1)*mpiCart.col;
            
            // firstRow, firstCol:
            // -> non-zero value of local grid is stored in (firstRow, firstCol)
            firstRow = (mpiCart.row==0) ? 1 : 0;
            firstCol = (mpiCart.col==0) ? 1 : 0;
                         
            // types to send rows and cols
            // MpiRow = MPI::DOUBLE.Create_contiguous(n+1); 
            MpiRow = MPI::DOUBLE.Create_contiguous(n+2); 
            MpiRow.Commit();
                
            MpiCol = MPI::DOUBLE.Create_vector(m+1, 1, grid.leadingDimension()); 
            MpiCol.Commit();
        }
        
        void
        updateBoundary()
        {
            MPI::Request req[8];
            MPI::Status  stat[8];
        
            req[ 0] = mpiCart.comm.Isend(&grid(0,n), 1, MpiCol, mpiCart.neighbors[East],  0);  
            req[ 1] = mpiCart.comm.Isend(&grid(m,0), 1, MpiRow, mpiCart.neighbors[South], 0);    
            req[ 2] = mpiCart.comm.Isend(&grid(0,0), 1, MpiCol, mpiCart.neighbors[West],  0);  
            req[ 3] = mpiCart.comm.Isend(&grid(0,0), 1, MpiRow, mpiCart.neighbors[North], 0);    

            req[ 4] = mpiCart.comm.Irecv(&grid(  0,n+1), 1, MpiCol, mpiCart.neighbors[East],  0);
            req[ 5] = mpiCart.comm.Irecv(&grid(m+1,  0), 1, MpiRow, mpiCart.neighbors[South], 0);
            req[ 6] = mpiCart.comm.Irecv(&grid(  0, -1), 1, MpiCol, mpiCart.neighbors[West],  0);
            req[ 7] = mpiCart.comm.Irecv(&grid( -1,  0), 1, MpiRow, mpiCart.neighbors[North], 0);
            
            MPI::Request::Waitall(8, req, stat);
        }
        
        MpiCart         mpiCart;
        int             N;
        int             m, n;
        int             m0, n0;
        int             firstRow, firstCol;
        Grid            grid;
        MPI::Datatype   MpiRow, MpiCol;
    };

    double
    norm(const DistributedVector &x)
    {
        double h = 1./(x.N+1);
        
        // compute (frobenius) norm
        // -> local norm
        double r = 0;
        for (int i=0; i<=x.m; ++i) {
            for (int j=0; j<=x.n; ++j) {
                r += x.grid(i,j)*x.grid(i,j);
            }
        }

        // -> global norm
        double R;
        x.mpiCart.comm.Allreduce(&r, &R, 1, MPI::DOUBLE, MPI::SUM);
        return sqrt(h*R);
    }

    double
    sum(const DistributedVector &x)
    {
        // -> local sum
        double s = 0;
        for (int i=0; i<=x.m; ++i) {
            for (int j=0; j<=x.n; ++j) {
                s += x.grid(i,j);
            }
        }

        // -> global sum
        double Sum;
        x.mpiCart.comm.Allreduce(&s, &Sum, 1, MPI::DOUBLE, MPI::SUM);
        return Sum;
    }

    void
    setRHS(DistributedVector &f)
    {
        double h = 1./(f.N+1);
        double x, y;
        for (int i=0; i<=f.m; ++i) {
            for (int j=0; j<=f.n; ++j) {
                x = (i+f.m0+0.5)*h;
                y = (j+f.n0+0.5)*h;

                f.grid(i,j) =  2*x+2*y-2;  //rhs
            }
        }
    }

    void
    setExactSolution(DistributedVector &e)
    {
        double h = 1./(e.N+1);

        for (int i=0; i<=e.m; ++i) {
            for (int j=0; j<=e.n; ++j) {
                double x = (i+e.m0+0.5)*h;
                double y = (j+e.n0+0.5)*h;

                e.grid(i,j) =  x*x/2 - x*x*x/3 + y*y/2 - y*y*y/3 - 1./6;
            }
        }
    }

    void
    gaussSeidelRedBlack(double omega, const DistributedVector &f, DistributedVector &u)
    {
        const DistributedVector::Grid &F = f.grid;
        DistributedVector::Grid       &U = u.grid;
        
        double c         = omega/4.;
        double hh        = 1./((u.N+1)*(u.N+1));

        if (u.mpiCart.row==0) {
            U(-1,_) = U(0,_);
        }
        if (u.mpiCart.row==u.mpiCart.numRows-1) {
            U(u.m+1,_) = U(u.m,_);
        }
        if (u.mpiCart.col==0) {
            U(_,-1) = U(_,0);
        }
        if (u.mpiCart.col==u.mpiCart.numCols-1) {
            U(_,u.n+1) = U(_,u.n);
        }

        // red nodes
        for (int i=0; i<=u.m; ++i) {
            for (int j=0; j<=u.n; j += 2) {
                U(i,j) = (1-omega)*U(i,j)
                       + c*(F(i,j)*hh + U(i-1,j) + U(i+1,j) + U(i,j-1) + U(i,j+1));
            }
            ++i;
            if(i<=u.m) {
                for (int j=1; j<=u.n; j += 2) {
                    U(i,j) = (1-omega)*U(i,j)
                       + c*(F(i,j)*hh + U(i-1,j) + U(i+1,j) + U(i,j-1) + U(i,j+1));
                }
            }
        }

        u.updateBoundary();

        // black nodes
        for (int i=0; i<=u.m; ++i) {
            for (int j=1; j<=u.n; j += 2) {
                U(i,j) = (1-omega)*U(i,j)
                       + c*(F(i,j)*hh + U(i-1,j) + U(i+1,j) + U(i,j-1) + U(i,j+1));        
            }
           ++i;
           if(i<=u.m) {
                for (int j=0; j<=u.n; j += 2) {
                    U(i,j) = (1-omega)*U(i,j)
                       + c*(F(i,j)*hh + U(i-1,j) + U(i+1,j) + U(i,j-1) + U(i,j+1));
                }
            }
        }

        u.updateBoundary();

        if (u.mpiCart.row==0) {
            U(-1,_) = U(0,_);
        }
        if (u.mpiCart.row==u.mpiCart.numRows-1) {
            U(u.m+1,_) = U(u.m,_);
        }
        if (u.mpiCart.col==0) {
            U(_,-1) = U(_,0);
        }
        if (u.mpiCart.col==u.mpiCart.numCols-1) {
            U(_,u.n+1) = U(_,u.n);
        }

        int    rhh = (u.N+1)*(u.N+1);
        double s   = sum(u)/rhh;

        for (int i=0; i<=u.m; ++i) {
            for (int j=0; j<=u.n; ++j) {
                U(i,j) -= s;
            }
        }
    }

    void
    residualPoisson(const DistributedVector &f,
                    const DistributedVector &u,
                    DistributedVector &r)
    {
        int rhh = (f.N+1)*(f.N+1);

        const DistributedVector::Grid &F = f.grid;
        const DistributedVector::Grid &U = u.grid;
        DistributedVector::Grid       &R = r.grid;

        for (int i=0; i<=f.m; ++i) {
            for (int j=0; j<=f.n; ++j) {
                R(i,j) = F(i,j) - rhh*(4*U(i,j) -U(i+1,j) -U(i-1,j) -U(i,j+1) -U(i,j-1));
            }
        }
    }

    double
    errorNorm(const DistributedVector &e, const DistributedVector &u)
    {
        double h = 1./(e.N+1);
        double r = 0;

        for (int i=0; i<=e.m; ++i) {
            for (int j=0; j<=e.n; ++j) {
                r += (e.grid(i,j)-u.grid(i,j))*(e.grid(i,j)-u.grid(i,j));
            }
        }

        double R;
        u.mpiCart.comm.Allreduce(&r, &R, 1, MPI::DOUBLE, MPI::SUM);
        return sqrt(h*R);
    }

    void
    printStat(int k,
              const DistributedVector &r,
              const DistributedVector &u,
              const DistributedVector &e)
    {
        double rNorm = norm(r);
        double eNorm = errorNorm(e,u);

        if ((r.mpiCart.row==0) && (r.mpiCart.col==0)) {
            cout.width(3);
            cout << k << ") | ";

            cout.precision(12);
            cout.setf(std::ios::fixed);
            cout.width(17);
            cout << rNorm << " | ";

            cout.precision(12);
            cout.setf(std::ios::fixed);
            cout.width(16);
            cout << eNorm << " | " << endl;
        }
    }

    void
    write(const DistributedVector &v, const string &name)
    {
        ostringstream s;
        s << name << "_local_" << v.mpiCart.row << "_" << v.mpiCart.col << ".dat";
        ofstream out(s.str().c_str());
        out << v.N << " " << v.N << endl;
        out << v.m-v.firstRow+1 << " " << v.n-v.firstCol+1<< endl; 
        out << v.grid(_(v.firstRow, v.m),_(v.firstCol, v.n));
    }

    void
    read(DistributedVector::Grid &grid, const string &name, int numProcX, int numProcY)
    {
        int row = 1;
        for (int px=0; px<numProcX; ++px) {
            int col = 1;
            for (int py=0; py<numProcY; ++py) {
                ostringstream s;
                s << name << "_local_" << px << "_" << py << ".dat";
                ifstream in(s.str().c_str());
                int N, M;
                in >> N >> M;
                if ((px==0) && (py==0)) {
                    grid.resize(N+2,M+2,0,0);
                }
                int r; int c;
                in >> r; in >> c;
                for (int i=0; i<r; ++i) {
                    for (int j=0; j<c; ++j) {
                        in >> grid(row+i,col+j);
                    }
                }
                col += c;
                if (py==numProcY-1) {
                    row += r;
                }
            }
        }
    }

    //------------------------------------------------------------------------------

    long l;// = 3;
    int  n;//= (1<<l) - 1;  // n = 2^l -1

    DistributedVector  u, f, r;

    void
    initGS(const MpiCart &mpiCart)
    {
        typedef DistributedVector Vec;
            u = Vec(mpiCart, n+1);
            f = Vec(mpiCart, n+1);
            r = Vec(mpiCart, n+1);
    }


    //------------------------------------------------------------------------------


    int
main(int argc, char *argv[])
{
    double t1,t2,t3,t4,t5,t6;
    t1 = clock();
    MPI::Init(argc, argv);
    int size = MPI::COMM_WORLD.Get_size();
    int p1, p2;
    t2 = clock();
    if(size==1) {
        p1 = 1;
        p2 = 1;
    }else if(size==2){
        p1 = 2;
        p2 = 1;
    }else{
        p1 = static_cast<int>(floor(sqrt(static_cast<double>(size))));

        if(p1%2!=0){
            p1 -= 1;
        }
        p2 = size/p1;
    }
    t3 = clock();
    MpiCart mpiCart(p1,p2);
    t4 = clock();
    if(mpiCart.rank == 0) {
        fstream in;
        in.open("infile.txt",ios::in);
        in   >> l;
        in.close();
        n = (1<<l) - 1;  // n = 2^l -1
//        cout << l << " " << n<< endl;
    }
    MPI_Bcast(&l, 1, MPI_INT , 0, MPI_COMM_WORLD);
    MPI_Bcast(&n, 1, MPI_INT , 0, MPI_COMM_WORLD);
    t5 = clock();
     /*u = new DistributedVector [l+1];
     f = new DistributedVector [l+1];
     r = new DistributedVector [l+1];*/

    initGS(mpiCart);

    DistributedVector e(mpiCart, n+1);
    setRHS(f);
    setExactSolution(e);


    residualPoisson(f, u, r);  
//    printStat(0, r, u, e);

    for (int k=1; k<=200*n; ++k) {
        gaussSeidelRedBlack(1., f, u);
        residualPoisson(f, u, r);  
//        printStat(k, r, u, e);
    }

    MPI::Finalize();
    t6 = clock();
    if(mpiCart.rank == 0) {
        fstream out;
        out.open("outfile.txt",ios::app|ios::out);
        out << (t2-t1 + t4-t3 + t6-t5)/CLOCKS_PER_SEC<< " " << n << " " << size<< endl;
        out.close();
   }
}
